from typing import List, Callable

import uvicorn
from fastapi import FastAPI
from pydantic import BaseSettings
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.middleware.cors import CORSMiddleware

from .config import ApiConfig
from .router import LeanRouter

router = LeanRouter()


class Modules:
    @classmethod
    def controllers(cls, modules: List[any]):
        return cls


class ApiServer:

    def __init__(self, r, configs):
        self.settings: ApiConfig = ApiConfig.load() if configs is None else configs
        self.host = ""
        self.port = 8001
        self.router = r
        self.app = FastAPI(
            title=self.settings.NAME,
            description=self.settings.DESCRIPTION,
            version=self.settings.VERSION,
            docs_url=f"{self.settings.BASE_URL}/docs",
            redoc_url=None,
            openapi_url=f"{self.settings.BASE_URL}/openapi.json"
        )

    @classmethod
    def config(cls, configs: BaseSettings = None):
        """
        Class configuration set and create instance
        """
        _class = cls(router, configs)
        return _class

    def loads(self, modules: List[any]):
        return self

    def middleware(self, func: Callable[[Request, Response], None]):
        @self.app.middleware("http")
        async def appmiddleware(request: Request, call_next: RequestResponseEndpoint) -> Response:
            response = await call_next(request)
            func(request, response)
            return response

        return self

    def request_middleware(self, func: Callable[[Request], None]):
        @self.app.middleware("http")
        async def appmiddleware(request: Request, call_next: RequestResponseEndpoint) -> Response:
            func(request)
            response = await call_next(request)
            return response

        return self

    def server(self):
        """
        Configuration server

        ApiServer.config().loads().server().start()
        """
        if self.settings.BACKEND_CORS_ORIGINS:
            self.app.add_middleware(
                CORSMiddleware,
                allow_origins=[str(origin) for origin in self.settings.BACKEND_CORS_ORIGINS],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )
        self.app.include_router(router)
        return self

    def start(self, host="127.0.0.1", port=8001):
        """
        Server will be start with uvicorn
        if host and post not defined then server use default values
        """
        self.host = host
        self.port = port
        print(f"[Host]: http://{self.host}:{self.port}")
        print(f"[Swagger]: http://{self.host}:{self.port}/{self.settings.BASE_URL.lstrip('/').rstrip('/')}/docs")
        print("*" * 87)
        uvicorn.run(self.app, host=self.host, port=self.port)



# -*- coding: utf-8 -*-
import secrets

from typing import Optional, Dict, Any, List, Union
from pydantic import (
    BaseSettings,
    Field, EmailStr, validator, AnyHttpUrl, HttpUrl,
)


class ApiConfig(BaseSettings):
    """
    CORS_ORIGINS    : Cors origins
    BASE_URL    : Swagger base url
    NAME        : Application Path
    DESCRIPTION : Application Description
    VERSION     : Application Version

    """
    CORS_ORIGINS: List[str] = ["*"]
    DOC_URL: str = "/base"
    NAME: str = "LEAN API"
    DESCRIPTION: str = ""
    VERSION: str = "0.0.0"

    @classmethod
    def load(cls):
        return cls()



import inspect
from typing import Any, Callable, List, Type, TypeVar, Union, get_type_hints

from fastapi import APIRouter, Depends
from pydantic.typing import is_classvar
from starlette.routing import Route, WebSocketRoute

T = TypeVar("T")
LEAN_CLASS_KEY = "__leanapi_class__"


def controller(path: str, router: APIRouter, depend: Callable[None] = None) -> Callable[[Type[T]], Type[T]]:
    def decorator(cls: Type[T]) -> Type[T]:
        return _controller(router, path, cls, [Depends(_call)] if _call else None)

    return decorator


def _controller(router: APIRouter, path: str, cls: Type[T], _call:Any) -> Type[T]:
    """
    Replaces any methods of the provided class `cls` that are endpoints of routes in `router` with updated
    function calls that will properly inject an instance of `cls`.
    """
    _init_controller(cls)
    cbv_router = APIRouter()
    function_members = inspect.getmembers(cls, inspect.isfunction)
    functions_set = set(func for _, func in function_members)
    cbv_routes = [
        route
        for route in router.routes
        if isinstance(route, (Route, WebSocketRoute)) and route.endpoint in functions_set
    ]
    for route in cbv_routes:
        router.routes.remove(route)
        _update_controller_route_endpoint_signature(cls, route)
        cbv_router.routes.append(route)
    router.include_router(cbv_router, prefix=f"/{path}", dependencies=_call, tags=[path])
    return cls


def _init_controller(cls: Type[Any]) -> None:
    """
    Idempotently modifies the provided `cls`, performing the following modifications:
    * The `__init__` function is updated to set any class-annotated dependencies as instance attributes
    * The `__signature__` attribute is updated to indicate to FastAPI what arguments should be passed to the initializer
    """
    if getattr(cls, LEAN_CLASS_KEY, False):
        return
    old_init: Callable[..., Any] = cls.__init__
    old_signature = inspect.signature(old_init)
    old_parameters = list(old_signature.parameters.values())[1:]  # drop `self` parameter
    new_parameters = [
        x for x in old_parameters if x.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
    ]
    dependency_names: List[str] = []
    for name, hint in get_type_hints(cls).items():
        if is_classvar(hint):
            continue
        parameter_kwargs = {"default": getattr(cls, name, Ellipsis)}
        dependency_names.append(name)
        new_parameters.append(
            inspect.Parameter(name=name, kind=inspect.Parameter.KEYWORD_ONLY, annotation=hint, **parameter_kwargs)
        )
    new_signature = old_signature.replace(parameters=new_parameters)

    def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
        for dep_name in dependency_names:
            dep_value = kwargs.pop(dep_name)
            setattr(self, dep_name, dep_value)
        old_init(self, *args, **kwargs)

    setattr(cls, "__signature__", new_signature)
    setattr(cls, "__init__", new_init)
    setattr(cls, LEAN_CLASS_KEY, True)


def _update_controller_route_endpoint_signature(cls: Type[Any], route: Union[Route, WebSocketRoute]) -> None:
    """
    Fixes the endpoint signature for a cbv route to ensure FastAPI performs dependency injection properly.
    """
    old_endpoint = route.endpoint
    old_signature = inspect.signature(old_endpoint)
    old_parameters: List[inspect.Parameter] = list(old_signature.parameters.values())
    old_first_parameter = old_parameters[0]
    new_first_parameter = old_first_parameter.replace(default=Depends(cls))
    new_parameters = [new_first_parameter] + [
        parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY) for parameter in old_parameters[1:]
    ]
    new_signature = old_signature.replace(parameters=new_parameters)
    setattr(route.endpoint, "__signature__", new_signature)

    



import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "leanapi",
    version = "0.0.6",
    author = "developerkitchen python team",
    author_email = "topluluk@developerkitchen.dev",
    description = ("Lean Api is a class layer for FastApi"),
    license = "MIT",
    keywords = "api, leanapi, fastapi",
    url = "http://github.com/developerkitchentr/leanapi",
    packages=['leanapi'],
    long_description_content_type = "text/markdown",
    long_description= read('README.md'),
    classifiers=[
        "Development Status :: 1 - Planning",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
    install_requires=[
        'fastapi>=0.63.0',
        'uvicorn>=0.13.3'
        'python-dotenv',
        'pydantic>=1.7.2'
    ]
)
